import numbers
from typing import Dict, Union
import enum


class Utils:

    @staticmethod
    def encode_val(value) -> Union[str, numbers.Number]: 
        """
        Encodes the given value to a string.

        Args:
            value: The value to encode.

        Returns:
            str or numeric: The encoded value.
        """
        return str(value).lower() if isinstance(value, (bool, str)) else value

    @staticmethod
    def update_args(row, args, keys=['timestamps', 'radius']) -> Dict:
        """
        Update the given arguments.

        Args:
            row: The value to be taken.
            args: The dict to update.
            keys: The keys to update.

        Returns:
            dict: Updated arguments
        """
        args.update({key: row[args[key]] for key in keys if isinstance(args.get(key), str)})
        return args
    
    @staticmethod
    def update_args_concurrent(row, args, keys=['timestamps', 'radius']) -> Dict:
        """
        Update the given arguments in concurrent race.

        Args:
            row: The value to be taken.
            args: The dict to update.
            keys: The keys to update.

        Returns:
            dict: Updated arguments
        """
        args.update({key: getattr(row, args[key]) for key in keys if isinstance(args.get(key), str)})
        return args
    
    @staticmethod
    def find_ways(ways, threshold, nodes):
        """
        Find the most appropriate ways.

        Args:
            ways: The input ways.
            threshold: The threshold used to filter out the ways.
            nodes: The input nodes.

        Returns:
            list: Selected ways
        """
        selected_ways = []
        for way_id, way_data in ways.items():
            intersection_size = max(0.001, len(set(way_data['node']) & set(nodes)))
            way_node_count = len(set(way_data['node']))
            nodes_count = len(set(nodes))

            precision = intersection_size / nodes_count
            recall = intersection_size / way_node_count

            f_beta_score = (1 + intersection_size**2) * (precision * recall) / (intersection_size**2 * precision + recall)

            if f_beta_score > threshold:
                selected_ways.append(way_id)
        
        return selected_ways


class ServiceType(enum.Enum):
    """
    Enum class representing different service types.
    """
    tile = "tile"
    route = "route"
    tags = "tags"
    elevation = "elevation"

    @classmethod
    def isService(cls, value: str) -> bool:
        """
        Check if the given value is a valid service type.

        Args:
            value: The value to check.

        Returns:
            True if the value is a valid service type, False otherwise.
        """
        return any(value == member.name for member in cls)


import json
import numbers
import copy
from typing import Any, Dict, List, Optional, Union, Tuple
import concurrent.futures
import numpy as np
import pandas as pd
import folium

from pyspark.sql import DataFrame
import pyspark.sql.functions as F
from pyspark.sql.types import StructType, StructField, ArrayType, DoubleType, StringType, MapType


from .elevation import ElevationServerEndpoint
from .overpass import OverpassServerEndpoint
from .tile import TileServerEndpoint
from .routing import RoutingServerEndpoint
from .utils import Utils, ServiceType
from .config import *


class QMap:

    WAY_TAGS = ["highway", "surface", "lanes", "maxspeed", 'crossing', 'traffic_signals']
    TAG_SEQ_ACCURACY = 0.7
    TAG_BLOCKSET={"name", "wikidata"}

    def __init__(self, api: str = "private"):
        """
        Initialize a QMap object.

        Args:
            api: The API to use ("public" or "private").
        """
        assert api in host_cfg.keys(), "The value of \"api\" parameter must be either \"public\" or \"private\""
        self._services = {}
        self._registerAll(host_cfg[api])

    def service(self, service: str):
        """
        Get the service object associated with the specified service type.

        Args:
            service: The service type.

        Returns:
            The service object associated with the specified service type.
        """
        assert ServiceType.isService(service), f"No service {service} to use"
        return self._services[service]

    def _registerAll(self, hosts):
        """
        Register all service endpoints using the provided hosts.

        Args:
            hosts: The host configuration for different services.
        """
        self._services[ServiceType.tile.value] = TileServerEndpoint(default_tile=hosts[ServiceType.tile.value])
        self._services[ServiceType.route.value] = RoutingServerEndpoint(host=hosts[ServiceType.route.value])
        self._services[ServiceType.tags.value] = OverpassServerEndpoint(host=hosts[ServiceType.tags.value])
        self._services[ServiceType.elevation.value] = ElevationServerEndpoint(host=hosts[ServiceType.elevation.value])

    def snap_to_road(self, data: Any, input_col_name: Optional[str] = None, timestamps: Any = None, radius: Any = None, tidy: bool = True, details: bool = False):
        """
        Snap the coordinates to the road.

        Args:
            data: The input data.
            input_col_name: The name of the input column.
            timestamps: Optional timestamps corresponding to the input coordinates.
            radius: Standard deviation of GPS precision used for map matching. If applicable use GPS accuracy.
            tidy: Indicates whether to tidy the output.
            details: Indicates whether to include detailed information.

        Returns:
            The snapped coordinates or detailed response.
        """
        udf_structure = StructType([
                            StructField('coordinates', ArrayType(ArrayType(DoubleType()))),
                            StructField('tags', StringType() if self.WAY_TAGS == "all" else StructType([StructField(t, ArrayType(StringType())) for t in self.WAY_TAGS])),
                            StructField('elevation', ArrayType(DoubleType()))
                        ]) if details else ArrayType(ArrayType(DoubleType()))
        return self._process_input(data, input_col_name, self._snap_to_road, udf_structure, timestamps=timestamps,radius=radius, tidy=tidy, details=details)

    def _process_input(self, data, input_col_name, func, udf_structure, **args):
        """
        Process the input data based on its type and perform the specified function.

        Args:
            data: The input data.
            input_col_name: The input column name.
            func: The function to be applied.
            udf_structure: The structure of the user-defined function (UDF).
            **args: Additional arguments for the function.

        Returns:
            The processed data.
        """
        timestamps = args.get('timestamps')
        radius = args.get('radius')

        if isinstance(data, (list, np.ndarray)):
            return func(data, **args)
    
        elif isinstance(data, pd.DataFrame):
            if isinstance(input_col_name, list):
                if isinstance(timestamps, str):
                    args['timestamps'] = data[timestamps].values
                if isinstance(radius, str):
                    args['radius'] = data[radius].values

                return func(data[input_col_name].values, **args)

            # SEQUENTIAL TASKS WITH PANDAS
            # data[col_name_cgf['outputCol']] = data.apply(lambda row: func(row[input_col_name], **Utils.update_args(row, copy.deepcopy(args))), axis=1)

            # PARALLEL TASKS WITH PANDAS
            def process_row(row, input_col_name, args):
                return func(getattr(row, input_col_name), **Utils.update_args_concurrent(row, args))
            with concurrent.futures.ThreadPoolExecutor() as executor:
                results = list(executor.map(lambda row: process_row(row, input_col_name, copy.deepcopy(args)), data.itertuples(index=False)))
            data[col_name_cgf['outputCol']] = results

            return data
        
        elif isinstance(data, DataFrame):
            if isinstance(input_col_name, list):
                raise ValueError(f"qmap supports only Batch Pyspark Processing in the current version. Look at the qsmap docs for more details")
            
            input_cols = [F.col(input_col_name)]
            if isinstance(timestamps, str):
                input_cols.append(F.col(args.pop('timestamps')))
            if isinstance(radius, str):
                input_cols.append(F.col(args.pop('radius')))

            converted_args = {key: F.lit(value) for key, value in args.items()}
            udf_func = F.udf(func, udf_structure)

            return data.withColumn(col_name_cgf['outputCol'], udf_func(*[*input_cols, *list(converted_args.values())]))

        else:  
            raise ValueError(f"Unknown {type(data)} type of the input data. qmap supports List, numpy.array, pandas.DataFrame, pyspark.sql.DataFrame in the current version")

    def _snap_to_road(self,coordinates, timestamps, radius, tidy, details) -> Any:
        """
        Internal method to snap the coordinates to the road.

        Args:
            coordinates: The input coordinates.
            timestamps: Optional timestamps corresponding to the input coordinates.
            radius: Standard deviation of GPS precision used for map matching. If applicable use GPS accuracy.
            tidy: Indicates whether to tidy the output.
            details: Indicates whether to include detailed information.

        Returns:
            The snapped coordinates or detailed response.
        """
        response = self._services[ServiceType.route.value].match(coordinates=coordinates, geometry="geojson", tidy=tidy, annotations=details,timestamps=timestamps,radius=radius)
        road = np.concatenate([sublist['geometry']['coordinates'] for sublist in response]).tolist()

        if not details:
            #return np.concatenate(list(map(lambda item: item['geometry']['coordinates'], response)))
            return None if len(road) == 0 else road
        
        if len(road) == 0: 
            return {"coordinates": None, "tags": None, "elevation": None}

        detailed_response = {}
        detailed_response['coordinates'] = road

        nodes = list(str(node) for subset in response for leg in subset.get("legs", []) for node in leg.get("annotation", {}).get("nodes", []))
        detailed_response['tags'] = self._way2tags(nodes)
        if self.WAY_TAGS == "all" and detailed_response['tags']: 
            detailed_response['tags'] = json.dumps(detailed_response['tags'])

        detailed_response['elevation'] = self._get_elevation(detailed_response['coordinates'])
        return detailed_response

    def _way2tags(self, nodes: List[str]) -> Dict:
        """
        Get tags from ways and nodes.

        Args:
            nodes: The input nodes.

        Returns:
            The dict of tags.
        """
        tags={}
        chunk_size = 600
        for start in range(0, len(nodes), chunk_size):
            chunk_nodes = nodes[start:min(start+chunk_size, len(nodes))]
            response=self._services[ServiceType.tags.value].fetch(f"""node(id:{','.join(chunk_nodes)});<;out;""")['osm'] #{"osm": {'way': [], 'node': []}}

            response['way'] = response.get('way', [])
            response['way'] = response['way'] if isinstance(response['way'], list) else [response['way']]
            ways = {}
            for way in response.get('way', []):
                nd = way.get('nd', [])
                tg = way.get('tag', [])
                nds = [n['@ref'] for n in (nd if isinstance(nd, list) else [nd]) if n]
                tgs = {t['@k']:t['@v'] for t in (tg if isinstance(tg, list) else [tg]) if t}
                ways[way['@id']] = {"node": nds, "tag": tgs}

            for iter_acc in range(13):
                selected_ways = Utils.find_ways(ways, self.TAG_SEQ_ACCURACY - (self.TAG_SEQ_ACCURACY / 14) * iter_acc, chunk_nodes)
                if selected_ways: break

            for selected_way in selected_ways:
                for k, v in ways[selected_way]['tag'].items():
                    if not any(blacklisted in k for blacklisted in self.TAG_BLOCKSET):
                        tags.setdefault(k, []).append(v)

            response=self._services[ServiceType.tags.value].fetch(f"""node(id:{','.join(chunk_nodes)});out;""")['osm']
            response['node'] = response.get('node', [])
            response['node'] = response['node'] if isinstance(response['node'], list) else [response['node']]

            for node in response.get('node', []):
                node_tags = node.get('tag', [])
                for tag in (node_tags if isinstance(node_tags, list) else [node_tags]):
                    k, v = tag.get('@k', ''), tag.get('@v', '')
                    if not any(blacklisted in k for blacklisted in self.TAG_BLOCKSET):
                        tags.setdefault(k, []).append(v)

        return self._filter_tags(tags)
    
    def _filter_tags(self, tags: Dict) -> Dict:
        """
        Filter keys in a dictionary based on WAY_TAGS.

        Args:
            tags: The input tags.

        Returns:
            The filtered tags based on WAY_TAGS.
        """
        if self.WAY_TAGS == "all":
            return tags
        elif isinstance(self.WAY_TAGS, list):
            return {key: tags.get(key, "None") for key in self.WAY_TAGS}
        else:
            print(f"Unknown WAY_TAGS={self.WAY_TAGS}")
            return None
        
    def get_elevation(self, data: Any, input_col_name: Optional[Union[str, List[str]]] = None):
        """
        Get the elevation.

        Args:
            data: The input data.
            input_col_name: The name of the input column(s).

        Returns:
            The elevation value(s).
        """
        udf_structure = DoubleType() if isinstance(input_col_name, list) else ArrayType(DoubleType())
        return self._process_input(data, input_col_name, self._get_elevation, udf_structure)
    
    def _get_elevation(self, coordinates) -> Union[float, List[float]]:
        """
        Internal method to get the elevation.

        Args:
            coordinates: The input coordinates.

        Returns:
            The elevation value(s).
        """
        response = self._services[ServiceType.elevation.value].estimate(coordinates)
        return response[0] if len(response) == 1 else response
    
    def plot_trajectory(self, data, input_col_name: Optional[Union[str, List[str]]] = None,
                        weight: int = 4, color: str = "black", opacity: float = 0.6, popup: Any = None, tile: folium.Map = None) -> folium.Map:
        """
        Plot the trajectory.

        Args:
            data: The input data.
            input_col_name: The name of the input column(s).
            weight: The weight of the polyline.
            color: The color of the polyline.
            opacity: The opacity of the polyline.
            popup: The popup content.
            tile: The tile object.

        Returns:
            The plotted tile object.
        """
        
        if isinstance(data, list):
            data = np.array(data)
        elif isinstance(data, DataFrame):
            data = data.toPandas()

        if isinstance(data, pd.DataFrame):
            if isinstance(input_col_name, list):
                data = data[input_col_name].values
            elif input_col_name:
                data = data[input_col_name]
            else:
                data = data.values

        if isinstance(data, np.ndarray):
            if len(data.shape) == 2:
                data = pd.Series(np.expand_dims(data, axis=0).tolist())
            data = pd.Series(data.tolist())

        tolist = lambda element: list(element) if isinstance(element[0], numbers.Number) else [tolist(subval) for subval in element]
        points = np.array([point for sublist in tolist(data.values) for point in sublist]).reshape(-1, 2)

        if not tile:
            tile = self._services[ServiceType.tile.value].create_tile(location=points.mean(axis=0).tolist())
        tile.fit_bounds([points.min(axis=0).tolist(), points.max(axis=0).tolist()])
        folium.PolyLine(data, weight=weight, color=color, opacity=opacity, popup=popup).add_to(tile)
        return tile



import requests
import json
import numbers
from typing import List, Union
import numpy as np 
import polyline

from .utils import Utils


class RoutingServerEndpoint:

    RESPONSE_KEYS_CFG = {
        "match": "matchings",
        "route": "routes"
    }

    def __init__(
        self,
        host: str = "http://router.project-osrm.org",
        version: str = "v1",
        profile: str = "driving",
        timeout: float = 5,
        max_retries: int = 5,
        pool_connections: int = 10,
        pool_maxsize: int = 10,
    ):
        """
        Initializes a RoutingServerEndpoint object.

        Args:
            host (str): The host URL of the routing server. Defaults to "http://router.project-osrm.org".
            version (str): The version of the routing API. Defaults to "v1".
            profile (str): The routing profile to use. Defaults to "driving".
            timeout (float): The request timeout in seconds. Defaults to 5.
            max_retries (int): The maximum number of retries for failed requests. Defaults to 5.
            pool_connections (int): The number of pool connections. Defaults to 10.
            pool_maxsize (int): The maximum size of pool. Defaults to 10.

        Raises:
            AssertionError: If the timeout value is not a number.
            AssertionError: If the max_retries value is not an integer or less than 1.
        """
        assert isinstance(timeout, numbers.Number), "Invalid timeout value"
        assert isinstance(max_retries, int) and max_retries >= 1, "Invalid max_retries value"

        self.host = host
        self.version = version
        self.profile = profile
        self.timeout = timeout
        self.max_retries = max_retries
        self.pool_connections = pool_connections
        self.pool_maxsize = pool_maxsize

        self._session = requests.Session()
        self._adapter = requests.adapters.HTTPAdapter(
            pool_connections=self.pool_connections,
            pool_maxsize=self.pool_maxsize,
            max_retries=self.max_retries)
        self._session.mount("http://", self._adapter)
        self._session.mount("https://", self._adapter)

    def _decode_response(self, response, service: str, params: dict) -> List:
        """
        Decodes the response received from the routing server.

        Args:
            response: The response object.
            service (str): The routing service used.
            params (dict): The request parameters.

        Returns:
            list: The decoded response.

        Raises:
            ValueError: If the response does not contain a valid code or if the geometry decoding fails.
        """
        try:
            response = response.json()
            if ("code" not in response) or ("Ok" not in response["code"]):
                raise ValueError(f"No valid response. Response={response}")

            if self.RESPONSE_KEYS_CFG[service] in response:

                geometry_processing_method = "decode" if params["geometries"] in ("polyline", "polyline6") else "flip"

                for item in response[self.RESPONSE_KEYS_CFG[service]]:

                    geometry = item.get("geometry")

                    if geometry is None: 
                        raise ValueError(f"No geometry found to decode")

                    if geometry_processing_method == "decode":
                        item["geometry"] = {"coordinates": np.array(polyline.decode(geometry))}
                    else:
                        item["geometry"] = {"coordinates": np.flip(geometry["coordinates"], axis=1)}
            else:
                raise ValueError(f"No {service} service found to decode")
            
            return response[self.RESPONSE_KEYS_CFG[service]]
        except Exception as error:
            print(f"Exception in RoutingServerEndpoint: {error}")
            return [{"geometry": {"coordinates": []}}]

    def match(
        self,
        coordinates,
        steps: bool = False,
        overview: str = "full",
        geometry: str = "polyline6",
        timestamps: list = None,
        radius: list = None,
        annotations: Union[bool,str] = False,
        gaps: str = "ignore",
        tidy: bool = False,
        waypoints: list = None,
    ) -> List:
        """
        Finds the best match for a set of input coordinates.

        Args:
            coordinates : The input coordinates.
            steps (bool): Whether to return step-by-step instructions. Defaults to False.
            overview (str): The level of overview geometry to be returned. Defaults to "full".
            geometry (str): The type of geometry to use in the response. Defaults to "polyline".
            timestamps (list): Optional timestamps corresponding to the input coordinates.
            radius (list): Standard deviation of GPS precision used for map matching. If applicable use GPS accuracy.
            annotations (bool or str): Whether to include additional metadata in the response. Defaults to False.
            gaps (str): Allows the input track splitting based on huge timestamp gaps between points. Defaults to "ignore".
            tidy (bool): Allows the input track modification to obtain better matching quality for noisy tracks. Defaults to False.
            waypoints (list): Treats input coordinates indicated by given indices as waypoints in returned Match object.

        Returns:
            list: The decoded response from the routing server.
        """
        service = "match"
        params = {
            "steps": Utils.encode_val(steps),
            "overview": overview,
            "geometries": geometry,
            "timestamps": ";".join(map(str, timestamps)) if timestamps is not None else timestamps,
            "radiuses": ";".join(map(str, radius)) if radius is not None else radius,
            "annotations": Utils.encode_val(annotations),
            "gaps": gaps,
            "tidy": Utils.encode_val(tidy),
            "waypoints": waypoints,
        }

        url = f"{self.host}/{service}/{self.version}/{self.profile}/{';'.join(map(lambda coord: f'{coord[1]},{coord[0]}', coordinates))}"
        return self._decode_response(self._session.get(url, params=params, timeout=self.timeout), service, params)



import folium

class TileServerEndpoint:

    TILES_CFG = {
        "QS Klokantech Basic": 'https://map-tile.ds.questarauto.com/styles/klokantech-basic/{z}/{x}/{y}.png',
        "QS OSM Bright": 'https://map-tile.ds.questarauto.com/styles/osm-bright/{z}/{x}/{y}.png',
        "OpenStreetMap": "OpenStreetMap",
        "Stamen Terrain": "Stamen Terrain",
        "Stamen Toner": "Stamen Toner",
        "Stamen Watercolor": "Stamen Watercolor",
        "CartoDB positron": "CartoDB positron",
        "CartoDB dark_matter": "CartoDB dark_matter"
    }

    def __init__(self, default_tile: str = "OpenStreetMap"):
        """
        Initializes a TileServerEndpoint object.

        Args:
            default_tile (str): The default tile server to use. Defaults to "OpenStreetMap".
        """
        self.default_tile = default_tile

    def create_tile(
        self,
        tiles: str = None,
        location: list[float] = [31.4117, 35.0818],
        zoom_start: float = 9.0,
        **kwargs
    ) -> folium.Map:
        """
        Creates a tile map using Folium library.

        Args:
            tiles (str): The tile server to use. Defaults to None, which uses the default_tile value.
            location (list[float]): The geographic coordinates of the center of the map. Defaults to [31.4117, 35.0818] (Israel).
            zoom_start (float): The initial zoom level of the map. Defaults to 9.0.
            **kwargs: Additional keyword arguments accepted by the folium.Map function.

        Returns:
            folium.Map: A Folium map object.

        Raises:
            Exception: If the specified tiles server is not found in TILES_CFG dictionary.
            Exception: If any unexpected error occurs during map creation.
        """
        tiles = tiles or self.default_tile
        try:
            return folium.Map(
                tiles=self.TILES_CFG[tiles],
                attr=tiles,
                location=location,
                zoom_start=zoom_start,
                **kwargs
            )
        except KeyError:
            print(f"KeyError in TileServerEndpoint. No '{tiles}' tile to use")
            return folium.Map(tiles="OpenStreetMap")
        except Exception as error:
            print(f"Exception in TileServerEndpoint: {error}")
            return folium.Map(tiles="OpenStreetMap")


import requests
import numbers
from typing import List
from itertools import chain
import numpy as np


class ElevationServerEndpoint:
    def __init__(
        self,
        host: str = "https://api.opentopodata.org",
        version: str = "v1",
        timeout: float = 5,
        max_retries: int = 5,
        pool_connections: int = 10,
        pool_maxsize: int = 10,
    ):
        """
        Initializes an ElevationServerEndpoint object.

        Args:
            host (str): The host URL of the elevation server. Defaults to "https://api.opentopodata.org".
            version (str): The version of the elevation API. Defaults to "v1".
            timeout (float): The request timeout in seconds. Defaults to 5.
            max_retries (int): The maximum number of retries for failed requests. Defaults to 5.
            pool_connections (int): The number of pool connections. Defaults to 10.
            pool_maxsize (int): The maximum size of pool. Defaults to 10.

        Raises:
            AssertionError: If the timeout value is not a number.
            AssertionError: If the max_retries value is not an integer or less than 1.
        """
        assert isinstance(timeout, numbers.Number), "Invalid timeout value"
        assert isinstance(max_retries, int) and max_retries >= 1, "Invalid max_retries value"

        self.host = host
        self.version = version
        self.timeout = timeout
        self.max_retries = max_retries
        self.pool_connections = pool_connections
        self.pool_maxsize = pool_maxsize

        self._session = requests.Session()
        self._adapter = requests.adapters.HTTPAdapter(
            pool_connections=self.pool_connections,
            pool_maxsize=self.pool_maxsize,
            max_retries=self.max_retries)
        self._session.mount("http://", self._adapter)
        self._session.mount("https://", self._adapter)

    def _decode_response(self, response) -> List:
        """
        Decodes the response received from the elevation server.

        Args:
            response: The response object.

        Returns:
            list: The decoded response.

        Raises:
            Exception: If the response does not contain a valid status.
        """
        try:
            response = response.json()
            if ("status" not in response) or ("OK" not in response.get("status", {})):
                raise ValueError(f"No valid response. Response={response}")
            return [result["elevation"] for result in response["results"]]
        except Exception as error:
            print(f"Exception in ElevationServerEndpoint: {error}")
            return [None]

    def estimate(self, coordinates, dataset: str = "mapzen") -> List:
        """
        Estimates the elevation for the given coordinates using the specified dataset.

        Args:
            coordinates: The input coordinates.
            dataset (str): The dataset to use. Defaults to "mapzen".

        Returns:
            list: The decoded response from the elevation server.
        """
        if np.ndim(coordinates) == 1:
            coordinates = np.expand_dims(coordinates, axis=0)
        
        if len(coordinates) > 500:
            return list(chain.from_iterable(map(self.estimate, np.array_split(coordinates, 2))))

        params = {  
            "locations": "|".join(map(lambda coord: f"{coord[0]},{coord[1]}", coordinates))
        }
        url = f"{self.host}/{self.version}/{dataset}"
        return self._decode_response(self._session.get(url, params=params, timeout=self.timeout))




import requests
import numbers
from typing import Dict
import xmltodict
import overpy

class OverpassServerEndpoint:

    def __init__(
        self, 
        host: str = "http://overpass-api.de/api/interpreter", 
        timeout: float = 30, 
        max_retries: int = 5,
        pool_connections: int = 10,
        pool_maxsize: int = 10,
    ):
        """
        Initializes an OverpassServerEndpoint object.

        Args:
            host (str): The host URL of the Overpass API. Defaults to "http://overpass-api.de/api/interpreter".
            timeout (float): The request timeout in seconds. Defaults to 30.
            max_retries (int): The maximum number of retries for failed requests. Defaults to 5.
            pool_connections (int): The number of pool connections. Defaults to 10.
            pool_maxsize (int): The maximum size of pool. Defaults to 10.

        Raises:
            AssertionError: If the timeout value is not a number.
            AssertionError: If the max_retries value is not an integer or less than 1.
        """
        assert isinstance(timeout, numbers.Number), "Invalid timeout value"
        assert isinstance(max_retries, int) and max_retries >= 1, "Invalid max_retries value"

        self.host = host
        self.timeout = timeout
        self.max_retries = max_retries
        self.pool_connections = pool_connections
        self.pool_maxsize = pool_maxsize
        
        self._session = requests.Session()
        self._adapter = requests.adapters.HTTPAdapter(
            pool_connections=self.pool_connections,
            pool_maxsize=self.pool_maxsize,
            max_retries=self.max_retries)
        self._session.mount("http://", self._adapter)
        self._session.mount("https://", self._adapter)

        self._api = overpy.Overpass(url=host, max_retry_count=max_retries, retry_timeout=timeout)

        
    def fetch(self, q: str) -> Dict:
        """
        Executes an Overpass query and returns the result.

        Args:
            q (str): The Overpass query string.

        Returns:
            Dict: The result of the query.

        Raises:
            ValueError: If the response does not contain a valid code or if the osm data decoding fails.
        """
        try:
            response = self._session.get(self.host, params={"data": q}, timeout=self.timeout)
            osm_data = xmltodict.parse(response.content).get("osm", {'way': [], 'node': []})
            if not osm_data: 
                raise ValueError("No osm data in service response")
            return {"osm": osm_data}
        except Exception as error:
            print(f"Exception in OverpassServerEndpoint: {error}")
            return {"osm": {'way': [], 'node': []}}

    def query(self, q: str) -> overpy.Result:
        """
        Executes an Overpass query and returns the result.

        Args:
            q (str): The Overpass query string.

        Returns:
            overpy.Result: The result of the query.
        """
        return self._api.query(q)




from .utils import ServiceType

col_name_cgf={
    'outputCol':'outputCol',
    'outputLatCol': 'outputLatCol',
    'outputLonCol': 'outputLonCol',
    'outputElevCol': 'outputElevCol',
    'inputLatLocCol': 'inputLatLocCol'
}

host_cfg={
    "public": {
        ServiceType.tile.value: "OpenStreetMap",
        ServiceType.route.value: "http://router.project-osrm.org",
        ServiceType.tags.value: "http://overpass-api.de/api/interpreter",
        ServiceType.elevation.value: "https://api.opentopodata.org"
    },
    "private": {
        ServiceType.tile.value: "QS OSM Bright",
        ServiceType.route.value: "https://map-routing.ds.questarauto.com",
        ServiceType.tags.value: "https://map-overpass.ds.questarauto.com/api/interpreter",
        ServiceType.elevation.value: "https://map-elevation.ds.questarauto.com"
    }
}


import setuptools

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name = 'qsmap', 
    packages = ['qsmap'],
    version='1.0.0',
    author='Yaroslav Mavliutov',
    author_email='yaroslavm@questarauto.com',
    description = 'Package provides functionality for working with geographical data, routing, and mapping',
    url='https://github.com/saferide-tech/QuestarMap',
    license='MIT',
    keywords = ['geo', 'routing', 'mapping'],
    install_requires=requirements,
)